/*
WMPVolKeys.cpp						Copyright  2004 by George A. Nelson

By:  Tony Nelson  on   4 Dec 04		

Description:  Intercept some Command Keys (those odd buttons across the top of
	some keyboards) and make them control the volume and mute of Windows Media
	Player.  This is my second attempt to control WMP volume and mute.  My
	first, using TweekUI and PostKeyTo, a little utility I wrote to post a
	keystroke to a running app, turned out to work only if the foreground app
	didn't already handle the Command Keys, which MS apps often do.  This
	version uses a Global Low Level Keyboard Hook to get the keys before the
	focused app does.  If the key is one we want, we eat it and post a Function
	Key to WMP if it is running (we search for it by class name).  If WMP is
	not running, we send the volume command keys, which affect the main volume
	settings.  The keys to intercept are configurable and are saved in the
	Registry.
	
	This program is derived from:

		Module:  DisableLowLevelKeys.cpp
		Notices: Written 2000 Jeffrey Richter

	A minimized app doesn't process menu accelerators (thank you, MVP Tim
	Robinson, for the info), as documented in TranslateAccelerator() and for
	WM_COMMAND, though it does do Window ("System") menu and unknown
	accelerators.  When WMP is minimized, we send a WM_COMMAND message; if
	not, we sent a WM_SYSKEYDOWN, as I expect that the published UI will change
	less often than the secret command values.

	As this app is a User Experience Extender (what would be a better name?),
	it should be able to minimize to the SysTray if the user wants it out of
	the way.  I use my new TrayButton.cp Window Behavior Subclass (what would
	be a better name?) to do this.

	Although GetKeyNameText() seemed like a good idea, it works from the scan
	code, while we work from the virtual key code; also it doesn't know the
	names of the Command Keys, which would have been the main reason to use it.

	-- Might want to "support" Apple's QuickTime Player, but it really sucks
	rocks.  Lower -> down arrow, raise -> up arrow, no way to mute.  The class
	name is the .exe path.  Ugly.  Stupid.  Metal.  Why I don't buy Apple
	products anymore.

	-- To support Win9x, we need a regular Global Keyboard Hook in a DLL.  I'm
	not quite ready to do this just now, though I know how.

*/

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

//#define UNICODE -- done in project settings

#include "..\\Common\\CmnHdr.h"				// Richter _Programming Applications..._
#include <WindowsX.h>
#include <stdlib.h>
#include <tchar.h>

#include "resource.h"
#include "..\\Common\\HyperLinks.h"
#include "..\\Common\\Utility.h"
#include "..\\TrayButton\\TrayButton.h"


/*	Registry key stuff:
*/
#define myGUID _T("{4484953A-C640-40b2-83F7-FBA3E7AB662D}")
#define myCompanyKey _T("TonyNelson") myGUID
#define myProductKey _T("WMPVolKeys")
#define vkInKeysName _T("VK")
#define trayMinimizeName _T("TrayMinimize")


/*	Volume Key type, used throughout and in the Registry VK data.
*/
typedef struct {
	BYTE key;
	BYTE mods;
} VolKey;

enum { vkiMute, vkiLower, vkiRaise, nVK };
enum { vkmShift = 0x01, vkmControl = 0x02 };


/*	Globals
*/
HINSTANCE			hInst;
static HHOOK		hhkLowLevelKybd;
static HWND			gMainHwnd;
static TCHAR		gDestAppClass[256];
static DWORD		gTrayMinimize = 0;

static VolKey		gvk[nVK] = {
	{ VK_BROWSER_SEARCH,    0 },
	{ VK_BROWSER_FAVORITES, 0 },
	{ VK_BROWSER_HOME,      0 },
};

VS_FIXEDFILEINFO	gVersInfo;


/*	Get keystrokes before applications do.  Unlike other hooks, this hook
	runs in the context of the setting app.
*/
   LRESULT CALLBACK
LowLevelKeyboardProc( int nCode, WPARAM wParam, LPARAM lParam )
{
	static BYTE sModifierState;
	BOOL fEatKeystroke = FALSE;

	if (nCode == HC_ACTION) {
		switch (wParam) {
		case WM_KEYDOWN:  case WM_SYSKEYDOWN:
		case WM_KEYUP:    case WM_SYSKEYUP: 
			PKBDLLHOOKSTRUCT p = (PKBDLLHOOKSTRUCT) lParam;
			BYTE vk = (BYTE) p->vkCode;
			WPARAM wmpKey;
			WPARAM wmpCmd;

			if( vk == gvk[vkiMute].key
			 && sModifierState == gvk[vkiMute].mods )
			{
				vk = VK_VOLUME_MUTE;
				wmpKey = VK_F8;
				wmpCmd = 0x4981;
				//fall through
			}
			else if( vk == gvk[vkiLower].key
			 && sModifierState == gvk[vkiLower].mods )
			{
				vk = VK_VOLUME_DOWN;
				wmpKey = VK_F9;
				wmpCmd = 0x4980;
				//fall through
			}
			else if( vk == gvk[vkiRaise].key
			 && sModifierState == gvk[vkiRaise].mods )
			{
				vk = VK_VOLUME_UP;
				wmpKey = VK_F10;
				wmpCmd = 0x497F;
				//fall through
			}
			else if( vk == VK_SHIFT
			      || vk == VK_LSHIFT
				  || vk == VK_RSHIFT )
			{
				if( p->flags & LLKHF_UP )
					sModifierState &= ~vkmShift;
				else
					sModifierState |= vkmShift;
				break;
			}
			else if( vk == VK_CONTROL
			      || vk == VK_LCONTROL
				  || vk == VK_RCONTROL )
			{
				if( p->flags & LLKHF_UP )
					sModifierState &= ~vkmControl;
				else
					sModifierState |= vkmControl;
				break;
			}
			else break;

			fEatKeystroke = true;

			HWND targetW = FindWindow( gDestAppClass, NULL );
			if( targetW ) {
				BOOL isIconic = IsIconic( targetW );
				switch( wParam ) {
				case WM_KEYDOWN:  case WM_SYSKEYDOWN:
					if( !isIconic )
						PostMessage( targetW, WM_SYSKEYDOWN, wmpKey, 0x00000001 );
					else
						PostMessage( targetW, WM_COMMAND, wmpCmd, 0 );
					break;

				case WM_KEYUP:    case WM_SYSKEYUP: 
					if( !isIconic )
						PostMessage( targetW, WM_SYSKEYUP,   wmpKey, 0xC0000001 );
					break;
				}
			}
			else if( vk == gvk[vkiMute].key	// Can't find window.
			      || vk == gvk[vkiLower].key
				  || vk == gvk[vkiRaise].key )
			{
				fEatKeystroke = false;		// Can't handle it -- endless recursion might result.
			}
			else {							// Send translated Virtual Key and hope for best.
				DWORD flags = 0;
				if( wParam == WM_KEYUP  ||  wParam == WM_SYSKEYUP )
					flags |= KEYEVENTF_KEYUP;
				keybd_event( vk, 0, flags, 0 );
			}
			break;
		}
	}

	return (fEatKeystroke) ? 1 : CallNextHookEx(hhkLowLevelKybd, nCode, wParam, lParam);
}


/*	Classify virtual keys according to some "sensible" scheme.  Use a big IF
	statement instead of a table so we can use names instead of table
	positions.  This is confusing enough already.  Some of it must be wrong.
*/
enum { vktUnk, vktNormal, vktModifier, vktMouse, vktFKey, vktInputMethod,
	vktCommand, vktMS, vktOEM };

	static int
VKClass( BYTE vk )
{
	int keyType = vktUnk;

	if( vk == VK_BACK
	 || vk == VK_TAB
	 || vk == VK_CLEAR
	 || vk == VK_RETURN
	 || vk == VK_PAUSE
	 || vk == VK_ESCAPE
	 || vk >= VK_SPACE    &&  vk <= VK_HELP
	 || vk >= '0'         &&  vk <= '9'
	 || vk >= 'A'         &&  vk <= 'Z'
	 || vk >= VK_NUMPAD0  &&  vk <= VK_DIVIDE
	 || vk == VK_NUMLOCK
	 || vk == VK_SCROLL
	 || vk >= VK_OEM_1    &&  vk <= VK_OEM_3
	 || vk >= VK_OEM_4    &&  vk <= VK_OEM_8
	 || vk == VK_OEM_AX
	 || vk == VK_OEM_102 )
	{
		keyType = vktNormal;
	}
	if( vk >= VK_SHIFT    &&  vk <= VK_MENU
	 || vk == VK_CAPITAL )
	{
		keyType = vktModifier;
	}
	else if(
		vk >= VK_LBUTTON  &&  vk <= VK_XBUTTON2	)
	{
		keyType = vktMouse;
	}
	else if(
		vk >= VK_F1       &&  vk <= VK_F24 )
	{
		keyType = vktFKey;
	}
	else if( 
		vk == VK_KANA
	 || vk >= VK_JUNJA    &&  vk <= VK_KANJI
	 || vk >= VK_CONVERT  &&  vk <= VK_MODECHANGE )
	{
		keyType = vktInputMethod;
	}
	else if(
		vk >= VK_BROWSER_BACK  &&  vk <= VK_LAUNCH_APP2 )
	{
		keyType = vktCommand;
	}
	else if(
		vk >= VK_LWIN     &&  vk <= VK_APPS
	 || vk == VK_SLEEP )
	{
		keyType = vktMS;
	}
	else if( 
		vk >= VK_OEM_FJ_JISHO  &&  vk <= VK_OEM_FJ_ROYA
	 || vk >= VK_CONVERT       &&  vk <= VK_MODECHANGE
	 || vk >= VK_ICO_HELP      &&  vk <= VK_ICO_00
	 ||	vk == VK_ICO_CLEAR
	 || vk >= VK_OEM_RESET     &&  vk <= VK_OEM_CLEAR )
	{
		keyType = vktOEM;
	}

	return keyType;
}


/*	Describe to the user the key to use.
*/
	static void
KeyDesc( VolKey & vk, TCHAR * desc, int size )
{
	desc[0] = 0;

	TCHAR modName[32];
	if( vk.mods & vkmControl ) {
		LoadString( hInst, IDS_VK_CONTROL, modName, chDIMOF(modName) );
		_tcsncat( desc, modName, size );
		_tcsncat( desc, _T("+"), size );
	}
	if( vk.mods & vkmShift ) {
		LoadString( hInst, IDS_VK_SHIFT, modName, chDIMOF(modName) );
		_tcsncat( desc, modName, size );
		_tcsncat( desc, _T("+"), size );
	}
	LoadString( hInst, IDS_VK_00+vk.key, modName, chDIMOF(modName) );
	_tcsncat( desc, modName, size );
}


/*	Message handlers for About dialog.
*/
	static BOOL
About_OnInitDialog( HWND hwnd, HWND hwndFocus, LPARAM lParam )
{
	CenterWindow( hwnd );

	/*	Put Product Version into dialog.
	*/
	HWND itemW = GetDlgItem( hwnd, IDC_ABOUTVERSION );
	TCHAR str[256], str1[128];
	GetWindowText( itemW, str1, chDIMOF(str1) );
	wsprintf( str, str1,
		HIWORD(gVersInfo.dwProductVersionMS),
		LOWORD(gVersInfo.dwProductVersionMS),
		HIWORD(gVersInfo.dwProductVersionLS),
		LOWORD(gVersInfo.dwProductVersionLS) );
	SetWindowText( itemW, str );

	/*	Subclass our mail link.
	*/
	ConvertStaticToHyperlink( hwnd, IDC_ABOUTMAILTO );

	return true;
}


	static void
About_OnCommand( HWND hwnd, int id, HWND hwndCtl, UINT codeNotify )
{
	switch( id ) {
	case IDOK:
		EndDialog( hwnd, id );
		break;

	case IDC_ABOUTMAILTO:
		if( codeNotify == STN_CLICKED ) {
			TCHAR str[256];
			LoadString( hInst, IDS_MAILTO, str, chDIMOF(str) );
			ShellExecute( hwnd, _T("open"), str, _T(""), _T(""), SW_SHOWNORMAL );
		}
		break;
	}
}


	static INT_PTR WINAPI
About_Proc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch( uMsg ) {
		chHANDLE_DLGMSG( hwnd, WM_INITDIALOG, About_OnInitDialog );
		chHANDLE_DLGMSG( hwnd, WM_COMMAND,    About_OnCommand );
	}

	return FALSE;
}


/*	Edit Key control subclass to intercepts keystrokes and set the edit field
	according to the key name.
*/
typedef struct {
	WNDPROC oldProc;
	VolKey vk;
} EditKeySubclassData;

static EditKeySubclassData eksd[nVK];

	LRESULT CALLBACK
EditKeySubclass( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	LRESULT result = 0;
	
	LONG eksdi = GetWindowLong(hwnd, GWL_USERDATA);
	if( (ULONG) eksdi >= nVK ) {
		static bool reported = false;
		if( !reported ) {
			MessageBox( hwnd, _T("Damaged EditKeySubclass.  Fooey."), _T("Internal Error"), MB_ICONSTOP );
			PostQuitMessage( 1 );
		}
		DefWindowProc( hwnd, uMsg, wParam, lParam );
	}
	EditKeySubclassData *eksdp = &eksd[ eksdi ];

	switch( uMsg ) {
	case WM_KEYDOWN:  case WM_SYSKEYDOWN:
		eksdp->vk.key  = (BYTE) wParam;
		eksdp->vk.mods = 0;
		if( eksdp->vk.key != VK_SHIFT    &&  GetKeyState(VK_SHIFT) < 0 )
			eksdp->vk.mods |= vkmShift;
		if( eksdp->vk.key != VK_CONTROL  &&  GetKeyState(VK_CONTROL) < 0 )
			eksdp->vk.mods |= vkmControl;

		TCHAR keyName[64];
		KeyDesc( eksdp->vk, keyName, chDIMOF(keyName) );
		SetWindowText( hwnd, keyName );
		Edit_SetSel( hwnd, 0, -1 );
		break;

	case WM_KEYUP:    case WM_SYSKEYUP:
	case WM_CHAR:
		break;

	case WM_GETDLGCODE:		// say we want all key processing (but Command Keys sneak past anyway)
		result = CallWindowProc( eksdp->oldProc, hwnd, uMsg, wParam, lParam );
		result &= ~ (LRESULT) DLGC_WANTCHARS;
		result |= DLGC_WANTALLKEYS | DLGC_WANTTAB;
		break;

	default:
		result = CallWindowProc( eksdp->oldProc, hwnd, uMsg, wParam, lParam );
	}

	return result;
}


/*	Message handlers for Set Keys dialog.
*/
		static void
	sSetKeys_OnInitEdit( HWND hwnd, int itemID, int vki )
	{
		HWND itemW = GetDlgItem( hwnd, itemID );
		
		EditKeySubclassData *eksdp = &eksd[vki];
		eksdp->vk = gvk[vki];
		eksdp->oldProc = SubclassWindow( itemW, EditKeySubclass );
		SetWindowLong( itemW, GWL_USERDATA, vki );
		
		TCHAR keyName[64];
		KeyDesc( eksdp->vk, keyName, chDIMOF(keyName) );
		SetWindowText( itemW, keyName );
	}

	static BOOL
SetKeys_OnInitDialog( HWND hwnd, HWND hwndFocus, LPARAM lParam )
{
	CenterWindow( hwnd );

	sSetKeys_OnInitEdit( hwnd, IDC_EDITMUTE,  vkiMute  );
	sSetKeys_OnInitEdit( hwnd, IDC_EDITLOWER, vkiLower );
	sSetKeys_OnInitEdit( hwnd, IDC_EDITRAISE, vkiRaise );
	
	return true;
}


		static bool
	sCheckVK( HWND hwnd, VolKey & vk, int itemID )
	{
		int vkClass = VKClass( vk.key );
		if( vk.key == 0 )
		{
			/*	I dont see how this can happen.
			*/
			MessageBeep( MB_ICONASTERISK );
			HWND itemW = GetDlgItem( hwnd, itemID );
			SetFocus( itemW );
			Edit_SetSel( itemW, 0, -1 );
			return false;
		}
		if( vkClass != vktUnk  &&  vkClass != vktCommand ) {
			TCHAR msg[256], str[140], str1[50], str2[64];
			LoadString( hInst, IDS_SETKEYSSETMSG, str, chDIMOF(str) );
			LoadString( hInst, IDS_VKCLASSNAME0+vkClass, str1, chDIMOF(str1) );
			KeyDesc( vk, str2, chDIMOF(str2) );
			wsprintf( msg, str, str1, str2 );
			LoadString( hInst, IDS_SETKEYSSETTTL, str, chDIMOF(str) );
			if( IDNO == MessageBox(hwnd, msg, str, MB_YESNO | MB_ICONQUESTION) ) {
				HWND itemW = GetDlgItem( hwnd, itemID );
				SetFocus( itemW );
				Edit_SetSel( itemW, 0, -1 );
				return false;
			}
		}
		
		return true;
	}

	static void
SetKeys_OnCommand( HWND hwnd, int id, HWND hwndCtl, UINT codeNotify )
{
	switch( id ) {
	case IDOK:
		if( !sCheckVK(hwnd, eksd[vkiMute].vk,  IDC_EDITMUTE )
		 || !sCheckVK(hwnd, eksd[vkiLower].vk, IDC_EDITLOWER)
		 || !sCheckVK(hwnd, eksd[vkiRaise].vk, IDC_EDITRAISE) )
		{
			break;
		}

		gvk[vkiMute]  = eksd[vkiMute].vk;
		gvk[vkiLower] = eksd[vkiLower].vk;
		gvk[vkiRaise] = eksd[vkiRaise].vk;

		/*	Save Command Keys in Registry.
		*/
		{
			HKEY rkProduct;
			DWORD disposition;
			LONG keyErr = RegCreateKeyEx( HKEY_CURRENT_USER,
				_T("Software\\") myCompanyKey _T("\\") myProductKey,
				0, NULL, 0, KEY_WRITE, NULL, &rkProduct, &disposition );
			if( keyErr == 0 ) {
				keyErr = RegSetValueEx( rkProduct, vkInKeysName, NULL, REG_BINARY,
					(BYTE *) gvk, sizeof gvk );
				keyErr = RegCloseKey( rkProduct );
			}
			CheckErrorMsgBox( hwnd, keyErr, IDS_SETKEYSFAILMSG, IDS_SETKEYSSETTTL );
		}
		// fall through to IDCANCEL

	case IDCANCEL:
		EndDialog( hwnd, id );
		break;
	}
}


	static INT_PTR WINAPI
SetKeys_Proc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch( uMsg ) {
		chHANDLE_DLGMSG( hwnd, WM_INITDIALOG, SetKeys_OnInitDialog );
		chHANDLE_DLGMSG( hwnd, WM_COMMAND,    SetKeys_OnCommand );

	case WM_APPCOMMAND:			// zap automatically processed Command Keys
		return TRUE;
	}

	return FALSE;
}


/*	Message handlers for our main window.
*/
	static void
SetMainKeysMsg( HWND hwnd )
{
	TCHAR msg[512], str[128], muteStr[64], lowerStr[64], raiseStr[64];
	LoadString( hInst, IDS_MAIN, str, chDIMOF(str) );
	KeyDesc( gvk[vkiMute],  muteStr,  chDIMOF(muteStr) );
	KeyDesc( gvk[vkiLower], lowerStr, chDIMOF(lowerStr) );
	KeyDesc( gvk[vkiRaise], raiseStr, chDIMOF(raiseStr) );
	wsprintf( msg, str, muteStr, lowerStr, raiseStr );
	HWND msgHwnd = GetDlgItem(hwnd, IDC_MAINMSG );
	SetWindowText( msgHwnd, msg );
}


	static BOOL
Dlg_OnInitDialog( HWND hwnd, HWND hwndFocus, LPARAM lParam )
{
	//chSETDLGICONS( hwnd, IDI_APPICON );
	SendMessage(hwnd, WM_SETICON, ICON_BIG,  (LPARAM) 
		LoadIcon(hInst, MAKEINTRESOURCE(IDI_APPICON)));
	SendMessage(hwnd, WM_SETICON, ICON_SMALL, (LPARAM) 
		(HICON) LoadImage( hInst, MAKEINTRESOURCE(IDI_APPICON), IMAGE_ICON,
		GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), 0));

	SetMainKeysMsg( hwnd );
	CheckDlgButton( hwnd, IDC_TRAYMINIMIZE, gTrayMinimize );
	return true;
}


	static void
Dlg_OnCommand( HWND hwnd, int id, HWND hwndCtl, UINT codeNotify )
{
	switch( id ) {
	case IDOK:
	case IDCANCEL:
		DestroyWindow( hwnd );
		break;

	case IDC_SETKEYS:
		UnhookWindowsHookEx( hhkLowLevelKybd );	
		DialogBoxParam( hInst, MAKEINTRESOURCE(IDD_SETKEYS), hwnd, SetKeys_Proc, 0 );
		hhkLowLevelKybd = SetWindowsHookEx( WH_KEYBOARD_LL, LowLevelKeyboardProc, hInst, 0 );
		SetMainKeysMsg( hwnd );
		break;

	case IDC_ABOUT:
		DialogBoxParam( hInst, MAKEINTRESOURCE(IDD_ABOUT), hwnd, About_Proc, 0 );
		break;
	
	case IDC_TRAYMINIMIZE:
		gTrayMinimize = IsDlgButtonChecked( hwnd, IDC_TRAYMINIMIZE ) == BST_CHECKED;
		SetTrayButton( hwnd, gTrayMinimize );

		/*	Save state in Registry.
		*/
		{
			HKEY rkProduct;
			DWORD disposition;
			LONG keyErr = RegCreateKeyEx( HKEY_CURRENT_USER,
				_T("Software\\") myCompanyKey _T("\\") myProductKey,
				0, NULL, 0, KEY_WRITE, NULL, &rkProduct, &disposition );
			if( keyErr == 0 ) {
				keyErr = RegSetValueEx( rkProduct, trayMinimizeName, NULL, REG_DWORD,
					(BYTE *) &gTrayMinimize, sizeof gTrayMinimize );
				keyErr = RegCloseKey( rkProduct );
			}
			CheckErrorMsgBox( hwnd, keyErr, IDS_TRAYMINIMIZEFAIL, IDS_ERRORTTL );
			break;
		}
	}
}


	static void
Dlg_OnDestroy( HWND hwnd )
{
	gMainHwnd = NULL;
	PostQuitMessage( 0 );
}


	static INT_PTR WINAPI
Dlg_Proc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch( uMsg ) {
		chHANDLE_DLGMSG( hwnd, WM_INITDIALOG,    Dlg_OnInitDialog );
		chHANDLE_DLGMSG( hwnd, WM_COMMAND,       Dlg_OnCommand );
		chHANDLE_DLGMSG( hwnd, WM_DESTROY,       Dlg_OnDestroy );
	}

	return FALSE;
}


	int APIENTRY
_tWinMain( HINSTANCE hInstExe, HINSTANCE, LPSTR pszCmdLine, int iCmdShow )	//_txxx from <tchar.h>
{
	hInst = hInstExe;

	chWindows9xNotAllowed();	// no LowLevelKeyboard hook in 9x

	/*	Get class name of Windows Media Player main window.
	*/
	LoadString( hInstExe, IDS_WMPCLASS, gDestAppClass, chDIMOF(gDestAppClass) );

	/*	Get Version info.
	*/
	{
		TCHAR name[MAX_PATH];
		GetModuleFileName( hInst, name, chDIMOF(name) );
		DWORD len = GetFileVersionInfoSize( name, NULL );
		void *data = malloc( len );
		GetFileVersionInfo( name, NULL, len, data );
		VS_FIXEDFILEINFO *info;
		UINT infoLen;
		VerQueryValue( data, _T("\\"), (LPVOID *) &info, &infoLen );
		gVersInfo = *info;
		free( data );
	}

	/*	Get Command Keys and Tray Minimize from Registry.  OK if missing or
		too small.
	*/
	HKEY rkProduct;
	LONG keyErr = RegOpenKeyEx( HKEY_CURRENT_USER,
		_T("Software\\") myCompanyKey _T("\\") myProductKey,
		0, KEY_READ, &rkProduct );
	if( keyErr == 0 ) {
		DWORD type;
		BYTE data[64];
		DWORD dataSize = sizeof data;
		if( 0 == RegQueryValueEx(rkProduct, vkInKeysName, NULL, &type, data, &dataSize)
		 && dataSize >= sizeof gvk )
		{
			memcpy( gvk, data, sizeof gvk );
		}
		dataSize = sizeof data;
		if( 0 == RegQueryValueEx(rkProduct, trayMinimizeName, NULL, &type, data, &dataSize)
		 && dataSize == sizeof gTrayMinimize )
		{
			memcpy( &gTrayMinimize, data, sizeof gTrayMinimize );
			gTrayMinimize = gTrayMinimize != 0;
		}
		keyErr = RegCloseKey( rkProduct );
	}

	/*	Run only one copy, so we install only one hook.
		
		The common way to detect multiple copies, by looking for an existing
		window, has a race condition if more than one copy is started at about
		the same time.  By creating the window first, we avoid that race
		condition, but then we need a modeless dialog and a message loop.

		A kernal object would be wrong, as hooks are per desktop, not per
		machine.
	*/
	gMainHwnd = CreateDialog( hInst, MAKEINTRESOURCE(IDD_MAIN), NULL, Dlg_Proc );
	if( !gMainHwnd )		// tampering?
		return 1;

	{
		TCHAR cn[256], wn[256];
		GetClassName( gMainHwnd, cn, chDIMOF(cn) );
		GetWindowText( gMainHwnd, wn, chDIMOF(wn) );
		HWND existsW = FindWindow( cn, wn );
		if( existsW == gMainHwnd )
			existsW = FindWindowEx( NULL, existsW, cn, wn );
		if( IsWindow(existsW) ) {
			SetForegroundWindow( existsW );
			DestroyWindow( gMainHwnd );
			return 0;	// quit!
		}
	}

	/*	Install the low-level keyboard hook.
	*/
	hhkLowLevelKybd = SetWindowsHookEx( WH_KEYBOARD_LL, LowLevelKeyboardProc, hInstExe, 0 );
	if( !hhkLowLevelKybd ) {
		ErrorMsgBox( gMainHwnd, GetLastError(), IDS_MAINNOHOOKMSG, IDS_ERRORTTL );
		return 1;
	}

	InitTrayButton( gMainHwnd, &iCmdShow, gTrayMinimize );
	ShowWindow( gMainHwnd, iCmdShow );
	
	MSG msg;
	BOOL msgErr;
	while( (msgErr = GetMessage(&msg, NULL, 0, 0)) != 0
	    && msgErr != -1 )
	{
		if( !gMainHwnd || !IsDialogMessage(gMainHwnd, &msg) ) { 
			TranslateMessage( &msg );
			DispatchMessage( &msg );
		}
	}

	/*	Keep the hook until we quit.
	*/
	if( hhkLowLevelKybd )
		UnhookWindowsHookEx( hhkLowLevelKybd );	
	
	return msg.wParam;
}
